from functools import reduce
import re
from typing import Callable, Iterable

from .rules import FilterRule
from .rules import (
    CLEAN_EXPLICIT_FILTER_RULES,
    FEATURE_FILTER_RULES,
    LIVE_FILTER_RULES,
    NORMALIZE_FEATURE_FILTER_RULES,
    PARODY_FILTER_RULES,
    REISSUE_FILTER_RULES,
    REMASTERED_FILTER_RULES,
    SUFFIX_FILTER_RULES,
    TRIM_SYMBOLS_FILTER_RULES,
    VERSION_FILTER_RULES,
    YOUTUBE_TRACK_FILTER_RULES,
)


__all__ = (
    "album_artist_from_artist",
    "filter_with_filter_rules",
    "fix_track_suffix",
    "normalize_feature",
    "remove_zero_width",
    "replace_nbsp",
    "remove_clean_explicit",
    "remove_live",
    "remove_reissue",
    "remove_remastered",
    "remove_version",
    "remove_parody",
    "remove_feature",
    "youtube",
)


FilterFunction = Callable[[str], str]


def album_artist_from_artist(text: str) -> str:
    """Generate Album Artist from Artist when "feat. Artist B" is present."""

    if " feat. " in text:
        return text.split(" feat. ")[0]
    return text


def filter_with_filter_rules(text: str, filter_rules: Iterable[FilterRule]) -> str:
    """Replace text in the given string according to given filter rules."""

    def reducer(text: str, filter_rule: FilterRule) -> str:
        return filter_rule.source.sub(
            filter_rule.target,
            text,
            count=filter_rule.count,
        )

    return reduce(reducer, filter_rules, text)


def fix_track_suffix(text: str) -> str:
    """Replace "Title - X Remix" suffix with "Title (X Remix) and similar"."""

    return filter_with_filter_rules(text, SUFFIX_FILTER_RULES)


def normalize_feature(text: str) -> str:
    """Generate normalized "feat. Artist B" text from [feat. Artist B] style."""

    return filter_with_filter_rules(text, NORMALIZE_FEATURE_FILTER_RULES)


def remove_clean_explicit(text: str) -> str:
    """Remove "Explicit" and "Clean"-like strings from the text."""

    return filter_with_filter_rules(text, CLEAN_EXPLICIT_FILTER_RULES)


def remove_feature(text: str) -> str:
    """Remove "feat"-like strings from the text."""

    return filter_with_filter_rules(text, FEATURE_FILTER_RULES)


def remove_live(text: str) -> str:
    """Remove "Live..."-like strings from the text."""

    return filter_with_filter_rules(text, LIVE_FILTER_RULES)


def remove_parody(text: str) -> str:
    """Remove "(Parody of "X" by Y)"-like strings from the text."""

    return filter_with_filter_rules(text, PARODY_FILTER_RULES)


def remove_reissue(text: str) -> str:
    """Remove "Re-issue"-like strings from the text."""

    return filter_with_filter_rules(text, REISSUE_FILTER_RULES)


def remove_remastered(text: str) -> str:
    """Remove "Remastered..."-like strings from the text."""

    return filter_with_filter_rules(text, REMASTERED_FILTER_RULES)


def remove_version(text: str) -> str:
    """Remove "(Single|Album|Mono version}"-like strings from the text."""

    return filter_with_filter_rules(text, VERSION_FILTER_RULES)


def remove_zero_width(text: str) -> str:
    """Remove zero-width characters from given string."""

    return re.sub("[\u200B-\u200D\uFEFF]", "", text)


def replace_nbsp(text: str) -> str:
    """Replace all non-breaking space symbols with a space symbol."""

    return text.replace("\u00a0", "\u0020")


def youtube(text: str) -> str:
    return filter_with_filter_rules(
        text,
        (
            *YOUTUBE_TRACK_FILTER_RULES,
            *TRIM_SYMBOLS_FILTER_RULES,
        ),
    )



from html import unescape

from .filter import MetadataFilter
from .functions import (
    album_artist_from_artist,
    fix_track_suffix,
    normalize_feature,
    remove_clean_explicit,
    remove_feature,
    remove_live,
    remove_parody,
    remove_reissue,
    remove_remastered,
    remove_version,
    youtube,
)


__all__ = (
    "make_youtube_filter",
    "make_remastered_filter",
    "make_spotify_filter",
    "make_amazon_filter",
    "make_tidal_filter",
)


def make_youtube_filter() -> MetadataFilter:
    """Get a filter with YouTube-related filter functions."""

    return MetadataFilter({"track": youtube})


def make_remastered_filter() -> MetadataFilter:
    """Get a filter that removes "Remastered"-like suffixes."""

    return MetadataFilter(
        {
            "track": remove_remastered,
            "album": remove_remastered,
        }
    )


def make_spotify_filter() -> MetadataFilter:
    """Get a filter with Spotify-related filter functions."""

    return MetadataFilter(
        {
            "track": (remove_remastered, remove_parody, fix_track_suffix, remove_live),
            "album": (
                remove_remastered,
                fix_track_suffix,
                remove_live,
                remove_reissue,
                remove_version,
            ),
        }
    )


def make_amazon_filter() -> MetadataFilter:
    """Get a filter with Amazon-related filter functions."""

    return MetadataFilter(
        {
            "artist": (normalize_feature,),
            "track": (
                remove_clean_explicit,
                remove_feature,
                remove_remastered,
                fix_track_suffix,
                remove_version,
                remove_live,
            ),
            "album": (
                unescape,
                remove_clean_explicit,
                remove_remastered,
                fix_track_suffix,
                remove_version,
                remove_live,
            ),
            "albumArtist": (
                normalize_feature,
                album_artist_from_artist,
            ),
        }
    )


def make_tidal_filter() -> MetadataFilter:
    """Get a filter with Tidal-related filter functions."""

    return MetadataFilter(
        {
            "track": (remove_remastered, fix_track_suffix, remove_version, remove_live),
            "album": (remove_remastered, fix_track_suffix, remove_version, remove_live),
        }
    )



from .functions import filter_with_filter_rules
from .opinionated_rules import SUFFIX_FILTER_RULES


def fix_track_suffix(text: str) -> str:
    """Replace "Title - X Remix" suffix with "Title (X Remix) and similar"."""

    return filter_with_filter_rules(text, SUFFIX_FILTER_RULES)



from .filter import MetadataFilter
from .functions import (
    remove_parody,
    remove_reissue,
    remove_remastered,
    remove_version,
)
from .opinionated_functions import fix_track_suffix


def make_spotify_filter() -> MetadataFilter:
    """
    Get a filter with an opinionated list of Spotify-related filter functions.
    Does not match upstream.
    """

    return MetadataFilter(
        {
            "track": (remove_remastered, remove_parody, fix_track_suffix),
            "album": (
                remove_remastered,
                fix_track_suffix,
                remove_reissue,
                remove_version,
            ),
        }
    )



import re

from .rules import FilterRule


SUFFIX_FILTER_RULES = (
    # "- X Remix" -> "(X Remix)" and similar
    FilterRule(
        source=re.compile(r"""-\s(.+?)\s((Re)?mix|edit|dub|mix|vip|version)$""", flags=re.IGNORECASE),
        target=r"""(\1 \2)""",
        count=1,
    ),
    FilterRule(
        source=re.compile(r"""-\s(Live|Remix|VIP|Instrumental)$""", flags=re.IGNORECASE),
        target=r"""(\1)""",
        count=1,
    ),
)


__all__ = [
    "SUFFIX_FILTER_RULES",
]



from collections import defaultdict
from collections.abc import Iterable
from functools import reduce
from typing import List, Mapping, MutableMapping, Union
from typing import Iterable as IterableType

from .functions import FilterFunction


__all__ = (
    "FilterFunction",
    "FilterFunctions",
    "FilterSet",
    "MetadataFilter",
)


FilterFunctions = IterableType[FilterFunction]
FilterSet = Mapping[str, Union[FilterFunction, FilterFunctions]]


def make_iterable(obj) -> Iterable:
    if isinstance(obj, Iterable):
        return obj
    return (obj,)


class MetadataFilter(object):
    def __init__(self, filter_set: FilterSet):
        self._merged_filter_set: MutableMapping[str, List[FilterFunction]] = defaultdict(lambda: [])
        self._append_filters(filter_set)

    def append(self, filter_set: FilterSet) -> "MetadataFilter":
        self._append_filters(filter_set)
        return self

    def extend(self, filter: "MetadataFilter") -> "MetadataFilter":
        self._append_filters(filter._merged_filter_set)
        return self

    def can_filter_field(self, field: str) -> bool:
        return field in self._merged_filter_set

    def get_fields(self) -> IterableType[str]:
        return self._merged_filter_set.keys()

    def filter_field(self, field: str, field_value: str) -> str:
        return self._filter_text(field_value, self._merged_filter_set[field])

    def _filter_text(self, text: str, filters: FilterFunctions) -> str:
        if not text:
            return text

        return reduce(lambda text, filter: filter(text), filters, text)

    def _append_filters(self, filter_set: FilterSet):
        for field, filters in filter_set.items():
            iterable_filters: FilterFunctions = make_iterable(filters)

            for filter in iterable_filters:
                if not callable(filter):
                    raise TypeError(
                        f"Invalid filter function: expected callable, got '{type(filter).__name__}'"
                    )
            self._merged_filter_set[field].extend(iterable_filters)



"""
DO NOT MODIFY THIS FILE DIRECTLY!
It is automatically generated by regen.sh
"""


from dataclasses import dataclass
import re


@dataclass(frozen=True)
class FilterRule(object):
    source: re.Pattern
    target: str = ""
    count: int = 0


CLEAN_EXPLICIT_FILTER_RULES = (
    # (Explicit) or [Explicit]
    FilterRule(
        source=re.compile(r"""\s[([]Explicit[)\]]""", flags=re.IGNORECASE),
        count=1,
    ),
    # (Clean) or [Clean]
    FilterRule(
        source=re.compile(r"""\s[([]Clean[)\]]""", flags=re.IGNORECASE),
        count=1,
    ),
)

FEATURE_FILTER_RULES = (
    # [Feat. Artist] or (Feat. Artist)
    FilterRule(
        source=re.compile(r"""\s[([]feat. .+[)\]]""", flags=re.IGNORECASE),
        count=1,
    ),
)

LIVE_FILTER_RULES = (
    # Track - Live
    # Track - Live at
    FilterRule(
        source=re.compile(r"""\s-\sLive(\s.+)?$"""),
        count=1,
    ),
    # Track (Live)
    FilterRule(
        source=re.compile(r"""\s[([]Live[)\]]$"""),
        count=1,
    ),
)

NORMALIZE_FEATURE_FILTER_RULES = (
    # [Feat. Artist] or (Feat. Artist) -> Feat. Artist
    FilterRule(
        source=re.compile(r"""\s[([](feat. .+)[)\]]""", flags=re.IGNORECASE),
        target=r""" \1""",
        count=1,
    ),
)

PARODY_FILTER_RULES = (
    # Party In the CIA (Parody of "Party In The U.S.A." by Miley Cyrus)
    FilterRule(
        source=re.compile(r"""\s\(Parody of ".*" by .*\)$"""),
        count=1,
    ),
    # White & Nerdy (Parody of "Ridin'" by Chamillionaire feat. Krayzie Bone)
    FilterRule(
        source=re.compile(r"""\s\(Parody of ".*" by .* feat\. .*\)$"""),
        count=1,
    ),
    # The Saga Begins (Lyrical Adaption of "American Pie")
    FilterRule(
        source=re.compile(r"""\s\(Lyrical Adaption of ".*"\)$"""),
        count=1,
    ),
)

REISSUE_FILTER_RULES = (
    # Album Title Re-issue
    FilterRule(
        source=re.compile(r"""\sRe-?issue$""", flags=re.IGNORECASE),
        count=1,
    ),
    # Album Title [Whatever Re-issue Whatever]
    FilterRule(
        source=re.compile(r"""\s\[.*?Re-?issue.*?\]""", flags=re.IGNORECASE),
        count=1,
    ),
    # Album Title (Whatever Re-issue Whatever)
    FilterRule(
        source=re.compile(r"""\s\(.*?Re-?issue.*?\)""", flags=re.IGNORECASE),
        count=1,
    ),
)

REMASTERED_FILTER_RULES = (
    # Ticket To Ride - Live / Remastered
    FilterRule(
        source=re.compile(r"""Live\s\/\sRemastered"""),
        target=r"""Live""",
        count=1,
    ),
    # Mothership (Remastered)
    # Let It Be (Remastered 2009)
    # How The West Was Won [Remastered]
    # Ride the Lightning (Deluxe Remaster)
    # ...And Justice For All (Remastered Deluxe Box Set)
    FilterRule(
        source=re.compile(r"""\s[([].*Re-?[Mm]aster(ed)?.*[)\]]$"""),
        count=1,
    ),
    # Outside The Wall - 2011 - Remaster
    # China Grove - 2006 Remaster
    # Easy Living - 2003 Remastered
    # Learning To Fly - 2001 Digital Remaster
    # Red Right Hand - 2011 Remastered Version
    FilterRule(
        source=re.compile(r"""\s-\s\d{4}(\s-)?\s.*Re-?[Mm]aster(ed)?.*$"""),
        count=1,
    ),
    # Here Comes The Sun - Remastered
    # 1979 - Remastered 2012
    # 1979 - Remastered Version
    FilterRule(
        source=re.compile(r"""\s-\sRe-?[Mm]aster(ed)?.*$"""),
        count=1,
    ),
    # Wish You Were Here [Remastered] (Remastered Version)
    FilterRule(
        source=re.compile(r"""\s\[Remastered\]\s\(Remastered\sVersion\)$"""),
        count=1,
    ),
)

SUFFIX_FILTER_RULES = (
    # "- X Remix" -> "(X Remix)" and similar
    FilterRule(
        source=re.compile(r"""-\s(.+?)\s((Re)?mix|edit|dub|mix|vip|version)$""", flags=re.IGNORECASE),
        target=r"""(\1 \2)""",
        count=1,
    ),
    FilterRule(
        source=re.compile(r"""-\s(Remix|VIP|Instrumental)$""", flags=re.IGNORECASE),
        target=r"""(\1)""",
        count=1,
    ),
)

TRIM_SYMBOLS_FILTER_RULES = (
    # Leftovers after e.g. (official video)
    FilterRule(
        source=re.compile(r"""\(+\s*\)+"""),
        count=1,
    ),
    # trim starting white chars and dash
    FilterRule(
        source=re.compile(r"""^[/,:;~\s"-]+"""),
        count=1,
    ),
    # trim trailing white chars and dash
    FilterRule(
        source=re.compile(r"""[/,:;~\s"-]+$"""),
        count=1,
    ),
    # remove multiple spaces
    FilterRule(
        source=re.compile(r"""\s{1,}"""),
        target=r""" """,
        count=1,
    ),
)

VERSION_FILTER_RULES = (
    # Love Will Come To You (Album Version)
    FilterRule(
        source=re.compile(r"""\s[([]Album Version[)\]]$"""),
        count=1,
    ),
    # I Melt With You (Rerecorded)
    # When I Need You [Re-Recorded]
    FilterRule(
        source=re.compile(r"""\s[([]Re-?recorded[)\]]$"""),
        count=1,
    ),
    # Your Cheatin' Heart (Single Version)
    FilterRule(
        source=re.compile(r"""\s[([]Single Version[)\]]$"""),
        count=1,
    ),
    # All Over Now (Edit)
    FilterRule(
        source=re.compile(r"""\s[([]Edit[)\]]$"""),
        count=1,
    ),
    # (I Can't Get No) Satisfaction - Mono Version
    FilterRule(
        source=re.compile(r"""\s-\sMono Version$"""),
        count=1,
    ),
    # Ruby Tuesday - Stereo Version
    FilterRule(
        source=re.compile(r"""\s-\sStereo Version$"""),
        count=1,
    ),
    # Pure McCartney (Deluxe Edition)
    FilterRule(
        source=re.compile(r"""\s\(Deluxe Edition\)$"""),
        count=1,
    ),
    # Ace of Spades (Expanded Edition)
    # Overkill (Expanded Bonus Track Edition)
    # On Parole (Expanded and Remastered)
    FilterRule(
        source=re.compile(r"""\s[([]Expanded.*[)\]]$"""),
        count=1,
    ),
    # Sound of White Noise - Expanded Edition
    FilterRule(
        source=re.compile(r"""\s-\sExpanded Edition$"""),
        count=1,
    ),
    # 6 Foot 7 Foot (Explicit Version)
    FilterRule(
        source=re.compile(r"""\s[([]Explicit Version[)\]]""", flags=re.IGNORECASE),
        count=1,
    ),
    # No Remorse (Bonus Track Edition)
    FilterRule(
        source=re.compile(r"""\s[([]Bonus Track Edition[)\]]""", flags=re.IGNORECASE),
        count=1,
    ),
    # Peace Sells...But Who's Buying (25th Anniversary)
    # Persistence of Time (30th Anniversary Remaster)
    FilterRule(
        source=re.compile(r"""\s[([]\d+th\sAnniversary.*[)\]]""", flags=re.IGNORECASE),
        count=1,
    ),
    # 6 Foot 7 Foot - Original
    FilterRule(
        source=re.compile(r"""\s-\sOriginal$""", flags=re.IGNORECASE),
        count=1,
    ),
    # California Love - Original Version
    # Personal Jesus - Original Single Version
    # Prince of the Moment - Original 7" Version
    # YMCA - Original Version 1978
    FilterRule(
        source=re.compile(r"""\s-\sOriginal.*Version(\s\d{4})?$""", flags=re.IGNORECASE),
        count=1,
    ),
)

YOUTUBE_TRACK_FILTER_RULES = (
    # Trim whitespaces
    FilterRule(
        source=re.compile(r"""^\s+|\s+$"""),
    ),
    # **NEW**
    FilterRule(
        source=re.compile(r"""\*+\s?\S+\s?\*+$"""),
        count=1,
    ),
    # [Whatever]
    FilterRule(
        source=re.compile(r"""\[[^\]]+\]"""),
        count=1,
    ),
    # (Whatever Version)
    FilterRule(
        source=re.compile(r"""\([^)]*version\)$""", flags=re.IGNORECASE),
        count=1,
    ),
    # Video extensions
    FilterRule(
        source=re.compile(r"""\.(avi|wmv|mpg|mpeg|flv)$""", flags=re.IGNORECASE),
        count=1,
    ),
    # (Lyrics Video)
    FilterRule(
        source=re.compile(r"""\(.*lyrics?\s*(video)?\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # ((Official)? (Track)? Stream)
    FilterRule(
        source=re.compile(r"""\((of+icial\s*)?(track\s*)?stream\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # ((Official)? (Music)? Video|Audio)
    FilterRule(
        source=re.compile(r"""\((of+icial\s*)?(music\s*)?(video|audio)\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # - (Official)? (Music)? Video|Audio
    FilterRule(
        source=re.compile(r"""-\s(of+icial\s*)?(music\s*)?(video|audio)$""", flags=re.IGNORECASE),
        count=1,
    ),
    # ((Whatever)? Album Track)
    FilterRule(
        source=re.compile(r"""\(.*Album\sTrack\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # (Official)
    FilterRule(
        source=re.compile(r"""\(\s*of+icial\s*\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # (1999)
    FilterRule(
        source=re.compile(r"""\(\s*[0-9]{4}\s*\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # (HD) / (HQ)
    FilterRule(
        source=re.compile(r"""\(\s*(HD|HQ)\s*\)$"""),
        count=1,
    ),
    # HD / HQ
    FilterRule(
        source=re.compile(r"""(HD|HQ)\s?$"""),
        count=1,
    ),
    # Video Clip Officiel / Video Clip Official
    FilterRule(
        source=re.compile(r"""(vid[\u00E9e]o)?\s?clip\sof+ici[ae]l""", flags=re.IGNORECASE),
        count=1,
    ),
    # Offizielles
    FilterRule(
        source=re.compile(r"""of+iziel+es\s*video""", flags=re.IGNORECASE),
        count=1,
    ),
    # Video Clip
    FilterRule(
        source=re.compile(r"""vid[\u00E9e]o\s?clip""", flags=re.IGNORECASE),
        count=1,
    ),
    # Clip
    FilterRule(
        source=re.compile(r"""\sclip""", flags=re.IGNORECASE),
        count=1,
    ),
    # Full Album
    FilterRule(
        source=re.compile(r"""full\s*album""", flags=re.IGNORECASE),
        count=1,
    ),
    # (Live)
    FilterRule(
        source=re.compile(r"""\(live.*?\)$""", flags=re.IGNORECASE),
        count=1,
    ),
    # | Something
    FilterRule(
        source=re.compile(r"""\|.*$""", flags=re.IGNORECASE),
        count=1,
    ),
    # Artist - The new "Track title" featuring someone
    FilterRule(
        source=re.compile(r"""^(|.*\s)"(.{5,})"(\s.*|)$"""),
        target=r"""\2""",
        count=1,
    ),
    # 'Track title'
    FilterRule(
        source=re.compile(r"""^(|.*\s)'(.{5,})'(\s.*|)$"""),
        target=r"""\2""",
        count=1,
    ),
    # (*01/01/1999*)
    FilterRule(
        source=re.compile(r"""\(.*[0-9]{1,2}\/[0-9]{1,2}\/[0-9]{2,4}.*\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # Sub Español
    FilterRule(
        source=re.compile(r"""sub\s*español""", flags=re.IGNORECASE),
        count=1,
    ),
    # (Letra)
    FilterRule(
        source=re.compile(r"""\s\(Letra\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # (En vivo)
    FilterRule(
        source=re.compile(r"""\s\(En\svivo\)""", flags=re.IGNORECASE),
        count=1,
    ),
    # Sub Español
    FilterRule(
        source=re.compile(r"""sub\s*español""", flags=re.IGNORECASE),
        count=1,
    ),
)


__all__ = (
    "FilterRule",
    "CLEAN_EXPLICIT_FILTER_RULES",
    "FEATURE_FILTER_RULES",
    "LIVE_FILTER_RULES",
    "NORMALIZE_FEATURE_FILTER_RULES",
    "PARODY_FILTER_RULES",
    "REISSUE_FILTER_RULES",
    "REMASTERED_FILTER_RULES",
    "SUFFIX_FILTER_RULES",
    "TRIM_SYMBOLS_FILTER_RULES",
    "VERSION_FILTER_RULES",
    "YOUTUBE_TRACK_FILTER_RULES",
)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="music-metadata-filter",
    version="2.0.0",
    author="Matthew Gamble",
    author_email="git@matthewgamble.net",
    description="A module for cleaning up artist, album, and song names.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/djmattyg007/music-metadata-filter",
    packages=setuptools.find_packages(include=("music_metadata_filter",)),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
    ],
    python_requires=">=3.7",
)
